"""Common worker library"""

import argparse
import inspect
import os
import re
import sys
from logging import debug, error
from typing import Any, Callable, Dict, Optional, Text, cast

import caep

import act.api

CONFIG_ID = "act"
CONFIG_NAME = "act.ini"


def parseargs(description: str, fact_arguments=False) -> argparse.ArgumentParser:
    """Parse arguments"""
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="{} ({})".format(description, worker_name()),
        epilog="""

  --config INI_FILE     Override default locations of ini file

    Arguments can be specified in ini-files, environment variables and
    as command line arguments, and will be parsed in that order.

    By default, configuration will be read from an ini file in /etc/{1}
    and ~/.config/{0}/{1} (or in $XDG_CONFIG_DIR if
    specified).

    Confiuration in "DEFAULT" section will be inherited by each secion.

    It is also possible to use environment variables for configuration.
    Environment variables should have the argument name in uppercase and
    "-" replaced with "_".

    E.g. set the CERT_FILE environment variable to configure the
    --cert-file option.

""".format(
            CONFIG_ID, CONFIG_NAME
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument("--http-timeout", type=int, default=120, help="Timeout")
    parser.add_argument("--proxy-string", help="Proxy to use for external queries")
    parser.add_argument(
        "--proxy-platform",
        action="store_true",
        help="Use proxy-string towards the ACT platform",
    )
    parser.add_argument(
        "--cert-file",
        help="Cerfiticate to add if you are behind a SSL/TLS interception proxy.",
    )
    parser.add_argument("--user-id", help="User ID")
    parser.add_argument("--act-baseurl", help="ACT API URI")
    parser.add_argument("--logfile", help="Log to file (default = stdout)")
    parser.add_argument("--loglevel", default="info", help="Loglevel (default = info)")
    parser.add_argument(
        "--http-header",
        help="Comma separated list of HTTP headers, e.g. "
        + "'HeaderA: val1, HeaderB:comma\\,val2",
    )

    parser.add_argument("--http-user", help="ACT HTTP Basic Auth user")
    parser.add_argument("--http-password", help="ACT HTTP Basic Auth password")

    if fact_arguments:
        parser.add_argument(
            "--output-format",
            dest="output_format",
            choices=["str", "json"],
            default="json",
            help="Output format for fact (default=json)",
        )
        parser.add_argument(
            "--access-mode",
            default=act.api.DEFAULT_ACCESS_MODE,
            choices=act.api.ACCESS_MODES,
            help="Specify default access mode used for all facts.",
        )
        parser.add_argument(
            "--organization", help="Specify default organization applied to all facts."
        )
        parser.add_argument(
            "--origin-name",
            dest="origin_name",
            help="Origin name. This name must be defined in the platform",
        )
        parser.add_argument(
            "--origin-id",
            dest="origin_id",
            help="Origin id. This must be the UUID of the origin in the platform",
        )
    return parser


def __mod_name(stack: inspect.FrameInfo) -> Text:
    """Return name of module from a stack ("_" is replaced by "-")"""
    mod = inspect.getmodule(stack[0])
    return os.path.basename(mod.__file__).replace(".py", "").replace("_", "-")


def worker_name() -> Text:
    """Return first external module that called this function, directly, or indirectly"""

    modules = [__mod_name(stack) for stack in inspect.stack() if __mod_name(stack)]
    return [name for name in modules if name != modules[0]][0]


def handle_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    """Wrapper for caep.handle_args where we set config_id and config_name"""
    args = caep.handle_args(parser, CONFIG_ID, CONFIG_NAME, worker_name())

    if args.http_header:
        # Convert comma separated list of http headers to dictionary
        headers = {}

        # Split on comma, unless they are escaped
        for header in re.split(r"(?<!\\),", args.http_header):
            if ":" not in header:
                raise act.api.base.ArgumentError(
                    f"No ':' in header, http header: {header}"
                )
            header_key, header_val = header.split(":", 1)
            header_key = header_key.strip().replace("\\,", ",")
            header_val = header_val.strip().replace("\\,", ",")
            headers[header_key] = header_val
        args.http_header = headers

    return cast(argparse.Namespace, args)


def fatal(message: Text, exit_code: int = 1) -> None:
    "Send error to error() and stderr() and exit with exit_code"
    sys.stderr.write(message.strip() + "\n")
    error(message.strip())
    sys.exit(exit_code)


def init_act(
    args: argparse.Namespace,
    object_formatter: Optional[Callable] = None,
    object_validator: Optional[Callable] = None,
) -> act.api.Act:
    """Initialize act api from arguments"""

    config = {
        "act_baseurl": args.act_baseurl,
        "user_id": args.user_id,
        "log_level": args.loglevel,
        "log_file": args.logfile,
        "log_prefix": worker_name(),
        # Provide defaults for optional arguments
        "origin_name": getattr(args, "origin_name", None),
        "origin_id": getattr(args, "origin_id", None),
        "access_mode": getattr(args, "access_mode", act.api.DEFAULT_ACCESS_MODE),
        "organization": getattr(args, "organization", None),
        "object_validator": object_validator,
        "object_formatter": object_formatter,
    }

    requests_kwargs: Dict[Text, Any] = {}

    if args.http_header:
        requests_kwargs["headers"] = args.http_header

    if args.http_user:
        requests_kwargs["auth"] = (args.http_user, args.http_password)

    if args.proxy_string and args.proxy_platform:
        requests_kwargs["proxies"] = {
            "http": args.proxy_string,
            "https": args.proxy_string,
        }

    if args.cert_file:
        requests_kwargs["verify"] = args.cert_file

    config["requests_common_kwargs"] = requests_kwargs

    api = act.api.Act(**config)

    if args.http_header:
        # Debug output of HTTP headers (must wait until act.api.Act() is initialized
        # so we have setup logging)
        debug("HTTP headers: %s", args.http_header)

    return api



import copy
import pickle
import random
import re
import uuid

import pytest

from act.api.re import TIMESTAMP_MATCH, UUID_MATCH
from act.api.schema import Field, Schema

# the Object/Fact and testdata here is a simplified version of the Object/Facts used in ACT
# The purpose is to only test core functionality of the Schema

DIRECTIONS = ["FactIsDestination", "FactIsSource", "BidirectionalFact"]


def random_hostname():
    """Return random (dummy) hostname"""
    return (
        "".join(
            [
                chr(random.randint(ord("a"), ord("z")))
                for i in range(1, random.randint(5, 12))
            ]
        )
        + ".com"
    )


class ObjectType(Schema):
    SCHEMA = [
        Field("name"),
        Field("id"),
    ]


class Object(Schema):
    SCHEMA = [
        Field("type", deserializer=ObjectType),
        Field("value"),
        Field("id"),
        Field("object", flatten=True),
        Field("direction"),
    ]

    def serialize(self, exclude_empty=True, to_camel_case=True):
        """Custom serializer for Object"""

        return {
            "object": {
                "id": self.id,
                "value": self.value,
                "type": {
                    "id": self.type.name,
                    "name": self.value,
                },
            },
            "direction": self.direction,
        }


class FactType(Schema):
    SCHEMA = [
        Field("name"),
        Field("id"),
    ]


class Fact(Schema):
    SCHEMA = [
        Field(
            "type", deserializer=FactType, serializer=lambda fact_type: fact_type.name
        ),
        Field("value"),
        Field("id", serializer=False),
        Field("timestamp", serializer=False),
        Field("objects", default=[], serialize_target="bindings", deserializer=Object),
    ]


fact_test_data = {
    "id": "00fb833d-3f74-4d70-9f00-d189eba6d038",
    "type": {"id": "c07d55c8-d976-4b06-841c-9d687c69bfe7", "name": "seenIn"},
    "value": "report",
    "timestamp": "2018-03-21T09:31:44.541Z",
    "objects": [],
}


object_test_data = {
    "object": {
        "id": "7dbf4d10-ff46-4d70-a681-ad08c2047d96",
        "type": {"id": "dbab2678-5110-405e-89b9-df6d5efe4a61", "name": "fqdn"},
        "value": "ukuoka.cloud-maste.com",
    },
    "direction": "FactIsDestination",
}


def __test_data():
    # Generate test data
    objects = []
    for _ in range(0, 5):  # Generate random facts
        obj = copy.copy(object_test_data)
        obj["id"] = str(uuid.uuid4())
        obj["object"]["value"] = random_hostname()
        obj["object"]["direction"] = random.choice(DIRECTIONS)
        objects.append(obj)

    fact = copy.copy(fact_test_data)
    fact["objects"] = objects

    return Fact(**fact)


def test_schema():
    f = __test_data()
    assert re.search(UUID_MATCH, f.id)
    assert re.search(UUID_MATCH, f.type.id)
    assert re.search(TIMESTAMP_MATCH, f.timestamp)

    # Facts should not have "bindings" as they are stored in "objects"
    with pytest.raises(AttributeError):
        # pylint: disable=pointless-statement
        f.bindings

    # Get first object
    obj = f.objects[0]

    # Objects should not have "object property" (the content from it should be
    # flattened into parrent)
    with pytest.raises(AttributeError):
        # pylint: disable=pointless-statement
        obj.object

    # When serialized, we should get back data in "object"
    assert isinstance(f.serialize()["bindings"][0]["object"], dict)


def test_pickle():
    """
    Test that we can pickle and unpickle a schema object and get the same schema
    """
    fact = __test_data()

    pickled_fact = pickle.dumps(fact)

    assert fact == pickle.loads(pickled_fact)


def test_set_get():
    f = __test_data()

    # type.name should be seenIn, both when referenced using dot
    # notation and when obtaining value through data dictionary
    assert f.type.name == "seenIn"
    assert f.data["type"].data["name"] == "seenIn"

    # Verify that we can set schema attributes directly using dot notation
    f.type.name = "mentions"

    # The name should then be the same when we reference it from the data dictionary
    assert f.data["type"].data["name"] == "mentions"



import copy
import json
import re
import types
from logging import error, info

import requests

from . import DEFAULT_ACCESS_MODE
from .re import UUID_MATCH
from .schema import Field, MissingField, Schema, schema_doc


class NotImplemented(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ArgumentError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ResponseError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ValidationError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


ERROR_HANDLER = {
    # Mapping of message templates provided in 412 errors from backend to
    # Exceptions that will be raised
    "fact.not.valid": lambda msg: ValidationError(
        "{message} ({field}={parameter})".format(**msg)
    ),
    "object.not.valid": lambda msg: ValidationError(
        "{message} ({field}={parameter})".format(**msg)
    ),
    "organization.not.exist": lambda msg: ValidationError(
        "{message} ({field}={parameter})".format(**msg)
    ),
}


def request(method, user_id, url, requests_common_kwargs=None, **kwargs):
    """Perform requests towards API

    Args:
        method (str):         POST|GET
        user_id (int):        Act user ID
        url (str):            Absolute URL for the endpoint
        **kwargs (keywords):  Additional options passed to requests json parameter
                              the following fields:"""

    if not requests_common_kwargs:
        requests_common_kwargs = {}

    requests_kwargs = copy.copy(requests_common_kwargs)

    # Apply common request arguments
    requests_kwargs.update(kwargs)

    # Make sure we have header as argument
    if "headers" not in requests_kwargs:
        requests_kwargs["headers"] = {}

    # Add User ID as header
    if user_id:
        requests_kwargs["headers"]["ACT-User-ID"] = str(user_id)

    try:
        res = requests.request(method, url, **requests_kwargs)
    except requests.exceptions.ConnectionError as e:
        raise ResponseError("Connection error {}".format(e))

    if res.status_code == 412:
        error_messages = res.json()["messages"]

        # Example output on object validation error:
        # {"responseCode": 412, "limit": 0, "count": 0, "messages": [{"type": "FieldError", "message": "Object did not pass validation against ObjectType.", "messageTemplate": "object.not.valid", "field": "objectValue", "parameter": "127.0.0.x", "timestamp": "2019-09-23T18:19:26.476Z"}], "data": null, "size": 0}

        # Raise ValidationError for 412/Object validation errors
        for msg in error_messages:
            msg_template = msg.get("messageTemplate")
            if msg_template in ERROR_HANDLER:
                raise ERROR_HANDLER[msg_template](msg)

        # All other, unhandled errors - log to error() and raise generic exception
        error("Request failed: {}, {}, {}".format(url, kwargs, res.status_code))
        raise ResponseError(res.text)

    elif res.status_code not in (200, 201):
        error("Request failed: {}, {}, {}".format(url, kwargs, res.status_code))
        raise ResponseError(
            "Unknown response error {}: {}".format(res.status_code, res.text)
        )

    try:
        return res.json()

    except json.decoder.JSONDecodeError:
        raise ResponseError(
            "Error decoding response {}: {}".format(res.status_code, res.text)
        )


class ActResultSet(object):
    """Represents a list of Act entries"""

    def __init__(self, response, deserializer, config=None):
        """Initialize result set
        Args:
            response (str):       JSON response from Act. This should include
                                  the following fields:
                                    - count: the number of entries fetched
                                    - limit: the limit in the query
                                    - responseCode: responseCode from the API
                                    - size: the total number of entries in the platform
                                    - data (array): array of entries

            deserializer (class): Deseralizer class"""

        if not isinstance(response["data"], list):
            raise ResponseError("Response should be list: {}".format(response["data"]))

        self.data = [deserializer(**d).configure(config) for d in response["data"]]

        self.size = response["size"]
        self.count = response["count"]
        self.limit = response["limit"]
        self.status_code = response["responseCode"]

    @property
    def complete(self):
        """Returns true if we have recieved all data that exists on the endpoint"""
        return self.size >= self.count

    def __call__(self, func, *args, **kwargs):
        """Call function on each data entry"""

        self.data = [getattr(item, func)(*args, **kwargs) for item in self.data]
        return self

    def __len__(self):
        """Returns the number of entries"""
        return len(self.data)

    def __getitem__(self, sliced):
        return self.data[sliced]

    def __str__(self):
        if not self.data:
            return "No result"
        return "\n".join(["{}".format(item) for item in self.data])

    def __bool__(self):
        """Return True for non empty result sets"""
        if self.size > 0:
            return True

        return False

    def __repr__(self):
        """
        Representation of result set
        """

        return repr(self.data)

    def __iter__(self):
        """Iterate over the entries"""
        return self.data.__iter__()


class Config(object):
    """Config object"""

    act_baseurl = None
    user_id = None
    requests_common_kwargs = {}

    def __init__(
        self,
        act_baseurl,
        user_id,
        requests_common_kwargs=None,
        origin_name=None,
        origin_id=None,
        access_mode=DEFAULT_ACCESS_MODE,
        organization=None,
        object_validator=None,
        object_formatter=None,
        strict_validator=False,
    ):
        """
        act_baseurl - url to ACT instance
        use_id - ACT user ID
        requests_common_kwargs - options that will be passed to requests when connecting to ACT api
        origin_name - ACT origin name that will be added to all facts where origin is not set
        origin_id - ACT origin id that will be added to all facts where origin is not set

        object_validator: function that should return True if object validates, otherwise None
        object_formatter: function that should return a formatted version of the type (e.g. lowercase)

        Only one of origin_name of origin_id must be specified.
        """

        if origin_id and not re.search(UUID_MATCH, origin_id):
            raise ArgumentError("origin_id id not a valaid UUID: {}".format(origin_id))

        self.act_baseurl = act_baseurl
        self.user_id = user_id
        self.requests_common_kwargs = requests_common_kwargs
        self.origin_name = origin_name
        self.origin_id = origin_id
        self.access_mode = access_mode
        self.organization = organization
        self.object_validator = object_validator
        self.object_formatter = object_formatter
        self.strict_validator = strict_validator


class ActBase(Schema):
    """Act object inheriting Schema, to support serializing and
    deserializing."""

    config = None

    SCHEMA = []

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(ActBase, self).__init__(*args, **kwargs)

    def configure(self, config):
        """Set config object"""

        self.config = config
        return self

    def api_request(self, method, uri, **kwargs):
        """Send request to API and update current object with result"""

        response = request(
            method,
            self.config.user_id,
            "{}/{}".format(self.config.act_baseurl, uri),
            self.config.requests_common_kwargs,
            **kwargs,
        )

        return response

    def api_post(self, uri, **kwargs):
        """Send POST request to API with keywords as JSON arguments"""

        return self.api_request("POST", uri, json=kwargs)

    def api_put(self, uri, **kwargs):
        """Send PUT request to API with keywords as JSON arguments"""

        return self.api_request("PUT", uri, json=kwargs)

    def api_delete(self, uri, params=None):
        """Send DELETE request to API
        Args:
            uri (str):     URI (relative to base url). E.g. "v1/factType"
            params (Dict): Parameters that are URL enncoded and sent to the API"""

        return self.api_request("DELETE", uri, params=params)

    def api_get(self, uri, params=None):
        """Send GET request to API
        Args:
            uri (str):     URI (relative to base url). E.g. "v1/factType"
            params (Dict): Parameters that are URL enncoded and sent to the API"""

        return self.api_request("GET", uri, params=params)

    def __eq__(self, other):
        "Equality operator"

        return hash(self) == hash(other)

    def __hash__(self):
        "__hash__ should be implemented on all derived classes"
        raise NotImplementedError(
            f"{self.__class__.__name__} is missing __hash__ method"
        )


class NameSpace(ActBase):
    """Namespace - serialized object specifying Namespace"""

    SCHEMA = [
        Field("name"),
        Field("id"),
    ]

    def __hash__(self):
        return hash(
            (
                self.__class__.__name__,
                self.name,
            )
        )


class Organization(ActBase):
    """Manage Organization"""

    SCHEMA = [
        Field("name"),
        Field("id"),
    ]

    def __hash__(self):
        return hash(
            (
                self.__class__.__name__,
                self.name,
            )
        )

    def serialize(self):
        # Return None for empty objects (non initialized objects)
        # otherwize return id or name
        return self.id or self.name or None


def origin_serializer(origin):
    # Return None for empty objects (non initialized origins)
    # otherwize return id or name
    if not origin:
        return None
    return origin.id or origin.name


class Origin(ActBase):
    """Manage Origin"""

    SCHEMA = [
        Field("name"),
        Field("id"),
        Field("namespace", serializer=False, deserializer=NameSpace),
        Field("organization", deserializer=Organization),
        Field("description"),
        Field("trust"),
        Field("type", serializer=False),
        Field("flags", serializer=False),
    ]

    def __hash__(self):
        return hash(
            (self.__class__.__name__, self.name, self.namespace, self.organization)
        )

    @schema_doc(SCHEMA)
    def __init__(self, *args, **kwargs):
        super(Origin, self).__init__(*args, **kwargs)

    def get(self):
        """Get Origin"""

        if not self.id:
            raise MissingField("Must have fact ID to get origin")

        origin = self.api_get("v1/origin/uuid/{}".format(self.id))["data"]
        self.data = {}
        self.deserialize(**origin)
        return self

    def add(self):
        """Add Origin"""
        params = self.serialize()

        origin = self.api_post("v1/origin", **params)["data"]

        # Empty data and load new result from response
        self.data = {}
        self.deserialize(**origin)

        info("Created origin: {}".format(self.name))

        return self

    def delete(self):
        """Delete Origin"""

        if not self.id:
            raise MissingField("Must have fact ID to delete origin")

        origin = self.api_delete("v1/origin/uuid/{}".format(self.id))["data"]
        self.data = {}
        self.deserialize(**origin)

        info("Deleted origin: {}".format(self.name))
        return self


class Comment(ActBase):
    """Namespace - serialized object specifying Namespace"""

    SCHEMA = [
        Field("comment"),
        Field("id"),
        Field("timestamp", serializer=False),
        Field("reply_to"),
        Field("origin", deserializer=Origin, serializer=False),
    ]

    def __hash__(self):
        return hash(
            (
                self.__class__.__name__,
                self.comment,
                self.timestamp,
            )
        )



"""Setup script for the python-act library module"""

from os import path

from setuptools import setup

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), "rb") as f:
    long_description = f.read().decode("utf-8")

setup(
    name="act-api",
    version="2.0.5",
    author="mnemonic AS",
    author_email="opensource@mnemonic.no",
    description="Python library to connect to the ACT rest API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    keywords="ACT, mnemonic",
    url="https://github.com/mnemonic-no",
    packages=["act.api", "act.api.libs"],
    namespace_packages=["act"],
    install_requires=["caep", "requests", "responses"],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "License :: OSI Approved :: ISC License (ISCL)",
    ],
)



""" Test for act helpers """

import pytest

import act.api


def test_add_uri_fqdn() -> None:  # type: ignore
    """Test for extraction of facts from uri with fqdn"""
    api = act.api.Act("", None, "error")

    uri = "http://www.mnemonic.no/home"

    facts = act.api.helpers.uri_facts(api, uri)

    assert len(facts) == 4
    assert (
        api.fact("componentOf")
        .source("fqdn", "www.mnemonic.no")
        .destination("uri", uri)
        in facts
    )
    assert (
        api.fact("componentOf").source("path", "/home").destination("uri", uri) in facts
    )
    assert api.fact("scheme", "http").source("uri", uri) in facts
    assert api.fact("basename", "home").source("path", "/home") in facts


def test_uri_should_fail() -> None:  # type: ignore
    """Test for extraction of facts from uri with ipv4"""
    api = act.api.Act("", None, "error")

    with pytest.raises(act.api.base.ValidationError):
        act.api.helpers.uri_facts(api, "http://")

    with pytest.raises(act.api.base.ValidationError):
        act.api.helpers.uri_facts(api, "www.mnemonic.no")

    with pytest.raises(act.api.base.ValidationError):
        act.api.helpers.uri_facts(api, "127.0.0.1")

    with pytest.raises(act.api.base.ValidationError):
        act.api.helpers.uri_facts(api, "http://localhost:test")


def test_add_uri_ipv4() -> None:  # type: ignore
    """Test for extraction of facts from uri with ipv4"""
    api = act.api.Act("", None, "error")

    uri = "http://127.0.0.1:8080/home"

    facts = act.api.helpers.uri_facts(api, uri)

    assert len(facts) == 5
    assert (
        api.fact("componentOf").source("ipv4", "127.0.0.1").destination("uri", uri)
        in facts
    )
    assert (
        api.fact("componentOf").source("path", "/home").destination("uri", uri) in facts
    )
    assert api.fact("scheme", "http").source("uri", uri) in facts
    assert api.fact("basename", "home").source("path", "/home") in facts
    assert api.fact("port", "8080").source("uri", uri) in facts


def test_add_uri_ipv6() -> None:  # type: ignore
    """Test for extraction of facts from uri with ipv4"""
    api = act.api.Act("", None, "error")

    uri = "http://[2001:67c:21e0::16]"

    facts = act.api.helpers.uri_facts(api, uri)

    assert len(facts) == 2
    assert api.fact("scheme", "http").source("uri", uri) in facts
    assert (
        api.fact("componentOf")
        .source("ipv6", "2001:067c:21e0:0000:0000:0000:0000:0016")
        .destination("uri", uri)
        in facts
    )


def test_add_uri_ipv6_with_port_path_query() -> None:  # type: ignore
    """Test for extraction of facts from uri with ipv6, path and query"""
    api = act.api.Act("", None, "error")

    uri = "http://[2001:67c:21e0::16]:8080/path?q=a"

    facts = act.api.helpers.uri_facts(api, uri)

    assert len(facts) == 6
    assert api.fact("scheme", "http").source("uri", uri) in facts
    assert (
        api.fact("componentOf")
        .source("ipv6", "2001:067c:21e0:0000:0000:0000:0000:0016")
        .destination("uri", uri)
        in facts
    )
    assert api.fact("port", "8080").source("uri", uri) in facts
    assert (
        api.fact("componentOf").source("path", "/path").destination("uri", uri) in facts
    )
    assert api.fact("basename", "path").source("path", "/path") in facts
    assert (
        api.fact("componentOf").source("query", "q=a").destination("uri", uri) in facts
    )


def test_ip_obj() -> None:
    """Test ip handling"""

    api = act.api.Act("", None, "error")

    assert act.api.helpers.ip_obj("2001:67c:21e0::16") == (
        "ipv6",
        "2001:067c:21e0:0000:0000:0000:0000:0016",
    )
    assert act.api.helpers.ip_obj("::1") == (
        "ipv6",
        "0000:0000:0000:0000:0000:0000:0000:0001",
    )
    assert act.api.helpers.ip_obj("127.0.0.1") == ("ipv4", "127.0.0.1")

    assert act.api.helpers.ip_obj("127.000.00.01") == ("ipv4", "127.0.0.1")

    with pytest.raises(ValueError):
        assert act.api.helpers.ip_obj("x.y.z") == ("ipv4", "x.y.x")

    with pytest.raises(ValueError):
        assert act.api.helpers.ip_obj("300.300.300.300") == ("ipv4", "x.y.x")

    assert api.fact("resolvesTo").source("fqdn", "localhost").destination(
        "ipv4", "127.0.0.1"
    ) == api.fact("resolvesTo").source("fqdn", "localhost").destination(
        *act.api.helpers.ip_obj("127.0.0.1")
    )
